@@ -1,87 +1,49 @@
-/* $Id: drawpix.c,v 3.14 1998/08/01 04:52:47 brianp Exp $ */
+/* $Id: drawpix.c,v 1.5.2.3 2000/04/15 03:18:44 brianp Exp $ */
 
 /*
  * Mesa 3-D graphics library
- * Version:  3.0
- * Copyright (C) 1995-1998  Brian Paul
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Library General Public
- * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Library General Public License for more details.
- *
- * You should have received a copy of the GNU Library General Public
- * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Version:  3.1
+ * 
+ * Copyright (C) 1999  Brian Paul   All Rights Reserved.
+ * 
+ * Permission is hereby granted, free of charge, to any person obtaining a
+ * copy of this software and associated documentation files (the "Software"),
+ * to deal in the Software without restriction, including without limitation
+ * the rights to use, copy, modify, merge, publish, distribute, sublicense,
+ * and/or sell copies of the Software, and to permit persons to whom the
+ * Software is furnished to do so, subject to the following conditions:
+ * 
+ * The above copyright notice and this permission notice shall be included
+ * in all copies or substantial portions of the Software.
+ * 
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
+ * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
+ * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
+ * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
+ * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
  */
 
 
-/*
- * $Log: drawpix.c,v $
- * Revision 3.14  1998/08/01 04:52:47  brianp
- * fixed conformance problems in GL_DEPTH_COMPONENT and GL_STENCIL_INDEX code
- *
- * Revision 3.13  1998/07/17 03:24:16  brianp
- * added Pixel.ScaleOrBiasRGBA field
- *
- * Revision 3.12  1998/06/28 22:32:22  brianp
- * fixed Mac compilation problem (Randy Frank and Miklos Fazekas)
- *
- * Revision 3.11  1998/06/01 00:38:13  brianp
- * fixed off-by-one bug in glDrawPixels clipping (Randy Frank)
- *
- * Revision 3.10  1998/05/31 23:50:36  brianp
- * cleaned up a few Solaris compiler warnings
- *
- * Revision 3.9  1998/04/01 02:58:52  brianp
- * applied Miklos Fazekas's 3-31-98 Macintosh changes
- *
- * Revision 3.8  1998/03/27 03:39:14  brianp
- * fixed G++ warnings
- *
- * Revision 3.7  1998/03/26 03:09:09  brianp
- * added GL_LUMINANCE and GL_LUMINANCE_ALPHA to gl_direct_DrawPixels()
- *
- * Revision 3.6  1998/03/25 01:56:03  brianp
- * fixed bug when Zoom.Y != 1.0 and top/bottom clipping needed (Randy Frank)
- *
- * Revision 3.5  1998/03/22 16:42:30  brianp
- * added CI->RGBA conversion to gl_direct_DrawPixels()
- *
- * Revision 3.4  1998/03/10 01:28:22  brianp
- * moved a few DEFARRY macros
- *
- * Revision 3.3  1998/02/08 20:22:14  brianp
- * LOTS of clean-up and rewriting
- *
- * Revision 3.2  1998/02/04 00:33:45  brianp
- * fixed a few cast problems for Amiga StormC compiler
- *
- * Revision 3.1  1998/02/01 22:16:34  brianp
- * include different headers (zooming)
- *
- * Revision 3.0  1998/01/31 20:51:03  brianp
- * initial rev
- *
- */
+
 
 
 #ifdef PC_HEADER
 #include "all.h"
 #else
+#ifndef XFree86Server
 #include <assert.h>
 #include <stdlib.h>
 #include <string.h>
+#else
+#include "GL/xf86glx.h"
+#endif
 #include "context.h"
 #include "drawpix.h"
 #include "feedback.h"
 #include "image.h"
 #include "macros.h"
+#include "mmath.h"
 #include "pixel.h"
 #include "span.h"
 #include "stencil.h"
@@ -111,10 +73,12 @@ GLboolean gl_direct_DrawPixels( GLcontext *ctx,
    GLubyte rgb[MAX_WIDTH][3];
    GLubyte rgba[MAX_WIDTH][4];
 
-   if (INSIDE_BEGIN_END(ctx)) {
-      gl_error( ctx, GL_INVALID_OPERATION, "glDrawPixels" );
-      return GL_TRUE;
-   }
+   ASSERT_OUTSIDE_BEGIN_END_AND_FLUSH_WITH_RETVAL(ctx, "glDrawPixels", 
+						  GL_FALSE);
+
+
+   if (ctx->RenderMode != GL_RENDER)
+      return GL_FALSE;
 
    if (!ctx->Current.RasterPosValid) {
       /* no-op */
@@ -153,7 +117,7 @@ GLboolean gl_direct_DrawPixels( GLcontext *ctx,
       GLint skipRows = unpack->SkipRows;
       GLint rowLength;
       GLdepth zSpan[MAX_WIDTH];  /* only used when zooming */
-      GLint zoomY0;
+      GLint zoomY0 = 0;
 
       if (unpack->RowLength > 0)
          rowLength = unpack->RowLength;
@@ -358,7 +322,8 @@ GLboolean gl_direct_DrawPixels( GLcontext *ctx,
                   assert(drawWidth < MAX_WIDTH);
                   gl_map_ci8_to_rgba(ctx, drawWidth, src, rgba);
                   (*ctx->Driver.WriteRGBASpan)(ctx, drawWidth, destX, destY,
-                                               rgba, NULL);
+                                               (const GLubyte (*)[4])rgba, 
+					       NULL);
                   src += rowLength;
                   destY++;
                }
@@ -433,7 +398,7 @@ static void draw_index_pixels( GLcontext *ctx, GLint x, GLint y,
       widthInBytes = width;
 
    /* Fragment depth values */
-   if (ctx->Depth.Test) {
+   if (ctx->Depth.Test || ctx->Fog.Enabled) {
       GLdepth zval = (GLdepth) (ctx->Current.RasterPos[2] * DEPTH_SCALE);
       for (i=0;i<width;i++) {
 	 zspan[i] = zval;
@@ -485,7 +450,8 @@ static void draw_index_pixels( GLcontext *ctx, GLint x, GLint y,
 	 GLubyte rgba[MAX_WIDTH][4];
          gl_map_ci_to_rgba( ctx, width, ispan, rgba );
          if (zoom) {
-            gl_write_zoomed_rgba_span( ctx, width, x, y, zspan, rgba, desty );
+            gl_write_zoomed_rgba_span( ctx, width, x, y, zspan, 
+				       (const GLubyte (*)[4])rgba, desty );
          }
          else {
             gl_write_rgba_span( ctx, width, x, y, zspan, rgba, GL_BITMAP );
@@ -521,6 +487,18 @@ static void draw_stencil_pixels( GLcontext *ctx, GLint x, GLint y,
    GLint i;
    const GLboolean zoom = ctx->Pixel.ZoomX!=1.0 || ctx->Pixel.ZoomY!=1.0;
 
+   if (image->Type != GL_BYTE &&
+       image->Type != GL_UNSIGNED_BYTE &&
+       image->Type != GL_SHORT &&
+       image->Type != GL_UNSIGNED_SHORT &&
+       image->Type != GL_INT &&
+       image->Type != GL_UNSIGNED_INT &&
+       image->Type != GL_FLOAT &&
+       image->Type != GL_BITMAP) {
+      gl_error( ctx, GL_INVALID_OPERATION, "glDrawPixels(stencil type)");
+      return;
+   }
+
    assert(image);
    assert(image->Format == GL_STENCIL_INDEX);
    assert(image->Type == GL_UNSIGNED_BYTE || image->Type == GL_BITMAP);
@@ -707,7 +685,7 @@ static void draw_depth_pixels( GLcontext *ctx, GLint x, GLint y,
          if (ctx->Visual->RGBAflag) {
             if (zoom) {
                gl_write_zoomed_rgba_span( ctx, width, x, y, zspan,
-                                          rgba, desty );
+                                          (const GLubyte (*)[4])rgba, desty );
             }
             else {
                gl_write_rgba_span( ctx, width, x, y, zspan, rgba, GL_BITMAP );
@@ -766,7 +744,7 @@ static void draw_rgba_pixels( GLcontext *ctx, GLint x, GLint y,
    height = image->Height;
 
    /* Fragment depth values */
-   if (ctx->Depth.Test) {
+   if (ctx->Depth.Test || ctx->Fog.Enabled) {
       /* fill in array of z values */
       GLdepth z = (GLdepth) (ctx->Current.RasterPos[2] * DEPTH_SCALE);
       for (i=0;i<width;i++) {
@@ -786,11 +764,12 @@ static void draw_rgba_pixels( GLcontext *ctx, GLint x, GLint y,
       /* General solution */
       GLboolean r_flag, g_flag, b_flag, a_flag, l_flag;
       GLuint components;
-      DEFARRAY(GLfloat, rf, MAX_WIDTH);
-      DEFARRAY(GLfloat, gf, MAX_WIDTH);
-      DEFARRAY(GLfloat, bf, MAX_WIDTH);
-      DEFARRAY(GLfloat, af, MAX_WIDTH);
       GLubyte rgba[MAX_WIDTH][4];
+      GLfloat  rf[MAX_WIDTH];
+      GLfloat  gf[MAX_WIDTH];
+      GLfloat  bf[MAX_WIDTH];
+      DEFARRAY(GLfloat,af,MAX_WIDTH);
+      CHECKARRAY(af,return);
 
       r_flag = g_flag = b_flag = a_flag = l_flag = GL_FALSE;
       switch (image->Format) {
@@ -892,19 +871,18 @@ static void draw_rgba_pixels( GLcontext *ctx, GLint x, GLint y,
 
 	 /* write to frame buffer */
          if (quickDraw) {
-            (*ctx->Driver.WriteRGBASpan)( ctx, width, x, y, rgba, NULL);
+            (*ctx->Driver.WriteRGBASpan)( ctx, width, x, y, 
+					  (const GLubyte (*)[4])rgba, NULL);
          }
          else if (zoom) {
-            gl_write_zoomed_rgba_span( ctx, width, x, y, zspan, rgba, desty );
+            gl_write_zoomed_rgba_span( ctx, width, x, y, zspan, 
+				       (const GLubyte (*)[4])rgba, desty );
          }
          else {
             gl_write_rgba_span( ctx, (GLuint) width, x, y, zspan, rgba, GL_BITMAP);
          }
       }
 cleanup:
-      UNDEFARRAY(rf);
-      UNDEFARRAY(gf);
-      UNDEFARRAY(bf);
       UNDEFARRAY(af);
    }
 }
@@ -916,10 +894,8 @@ cleanup:
  */
 void gl_DrawPixels( GLcontext* ctx, struct gl_image *image )
 {
-   if (INSIDE_BEGIN_END(ctx)) {
-      gl_error( ctx, GL_INVALID_OPERATION, "glDrawPixels" );
-      return;
-   }
+   ASSERT_OUTSIDE_BEGIN_END_AND_FLUSH(ctx, "glDrawPixels");
+
 
    if (gl_image_error_test( ctx, image, "glDrawPixels" ))
       return;
@@ -960,22 +936,17 @@ void gl_DrawPixels( GLcontext* ctx, struct gl_image *image )
    }
    else if (ctx->RenderMode==GL_FEEDBACK) {
       if (ctx->Current.RasterPosValid) {
-         GLfloat color[4], texcoord[4], invq;
-         color[0] = (GLfloat) ctx->Current.ByteColor[0] / 255.0;
-         color[1] = (GLfloat) ctx->Current.ByteColor[1] / 255.0;
-         color[2] = (GLfloat) ctx->Current.ByteColor[2] / 255.0;
-         color[3] = (GLfloat) ctx->Current.ByteColor[3] / 255.0;
-         invq = 1.0F / ctx->Current.TexCoord[3];
-         texcoord[0] = ctx->Current.TexCoord[0] * invq;
-         texcoord[1] = ctx->Current.TexCoord[1] * invq;
-         texcoord[2] = ctx->Current.TexCoord[2] * invq;
-         texcoord[3] = ctx->Current.TexCoord[3];
+         GLfloat color[4];
+	 GLfloat texcoord[4], invq;
+	 UBYTE_RGBA_TO_FLOAT_RGBA(color, ctx->Current.ByteColor);
+         invq = 1.0F / ctx->Current.Texcoord[0][3];
+         texcoord[0] = ctx->Current.Texcoord[0][0] * invq;
+         texcoord[1] = ctx->Current.Texcoord[0][1] * invq;
+         texcoord[2] = ctx->Current.Texcoord[0][2] * invq;
+         texcoord[3] = ctx->Current.Texcoord[0][3];
          FEEDBACK_TOKEN( ctx, (GLfloat) (GLint) GL_DRAW_PIXEL_TOKEN );
          gl_feedback_vertex( ctx,
-                             ctx->Current.RasterPos[0],
-                             ctx->Current.RasterPos[1],
-                             ctx->Current.RasterPos[2],
-                             ctx->Current.RasterPos[3],
+                             ctx->Current.RasterPos,
                              color, ctx->Current.Index, texcoord );
       }
    }
