@@ -1,35 +1,31 @@
-/* $Id: masking.c,v 3.1 1998/02/08 20:19:41 brianp Exp $ */
+/* $Id: masking.c,v 1.3 1999/10/08 09:27:11 keithw Exp $ */
 
 /*
  * Mesa 3-D graphics library
- * Version:  3.0
- * Copyright (C) 1995-1998  Brian Paul
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Library General Public
- * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Library General Public License for more details.
- *
- * You should have received a copy of the GNU Library General Public
- * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Version:  3.1
+ * 
+ * Copyright (C) 1999  Brian Paul   All Rights Reserved.
+ * 
+ * Permission is hereby granted, free of charge, to any person obtaining a
+ * copy of this software and associated documentation files (the "Software"),
+ * to deal in the Software without restriction, including without limitation
+ * the rights to use, copy, modify, merge, publish, distribute, sublicense,
+ * and/or sell copies of the Software, and to permit persons to whom the
+ * Software is furnished to do so, subject to the following conditions:
+ * 
+ * The above copyright notice and this permission notice shall be included
+ * in all copies or substantial portions of the Software.
+ * 
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
+ * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
+ * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
+ * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
+ * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
  */
 
 
-/*
- * $Log: masking.c,v $
- * Revision 3.1  1998/02/08 20:19:41  brianp
- * ColorMask is now GLubyte[4] instead of GLuint
- *
- * Revision 3.0  1998/01/31 20:58:46  brianp
- * initial rev
- *
- */
+
 
 
 /*
@@ -40,9 +36,15 @@
 #ifdef PC_HEADER
 #include "all.h"
 #else
+#ifndef XFree86Server
 #include <string.h>
+#include <stdio.h>
+#else
+#include "GL/xf86glx.h"
+#endif
 #include "alphabuf.h"
 #include "context.h"
+#include "enums.h"
 #include "macros.h"
 #include "masking.h"
 #include "pb.h"
@@ -54,10 +56,7 @@
 
 void gl_IndexMask( GLcontext *ctx, GLuint mask )
 {
-   if (INSIDE_BEGIN_END(ctx)) {
-      gl_error( ctx, GL_INVALID_OPERATION, "glIndexMask" );
-      return;
-   }
+   ASSERT_OUTSIDE_BEGIN_END_AND_FLUSH(ctx, "glIndexMask");
    ctx->Color.IndexMask = mask;
    ctx->NewState |= NEW_RASTER_OPS;
 }
@@ -67,15 +66,19 @@ void gl_IndexMask( GLcontext *ctx, GLuint mask )
 void gl_ColorMask( GLcontext *ctx, GLboolean red, GLboolean green,
                    GLboolean blue, GLboolean alpha )
 {
-   if (INSIDE_BEGIN_END(ctx)) {
-      gl_error( ctx, GL_INVALID_OPERATION, "glColorMask" );
-      return;
-   }
+   ASSERT_OUTSIDE_BEGIN_END_AND_FLUSH(ctx, "glColorMask");
+
+   if (MESA_VERBOSE & VERBOSE_API)
+      fprintf(stderr, "glColorMask %d %d %d %d\n", red, green, blue, alpha);
+
    ctx->Color.ColorMask[RCOMP] = red    ? 0xff : 0x0;
    ctx->Color.ColorMask[GCOMP] = green  ? 0xff : 0x0;
    ctx->Color.ColorMask[BCOMP] = blue   ? 0xff : 0x0;
    ctx->Color.ColorMask[ACOMP] = alpha  ? 0xff : 0x0;
 
+   if (ctx->Driver.ColorMask) 
+      ctx->Driver.ColorMask( ctx, red, green, blue, alpha );
+
    ctx->NewState |= NEW_RASTER_OPS;
 }
 
@@ -119,6 +122,9 @@ void gl_mask_rgba_pixels( GLcontext *ctx,
    GLuint i;
 
    (*ctx->Driver.ReadRGBAPixels)( ctx, n, x, y, dest, mask );
+   if (ctx->RasterMask & ALPHABUF_BIT) {
+      gl_read_alpha_pixels( ctx, n, x, y, dest, mask );
+   }
 
    for (i=0; i<n; i++) {
       rgba32[i] = (rgba32[i] & srcMask) | (dest32[i] & dstMask);
