@@ -1,77 +1,27 @@
-/* $Id: tritemp.h,v 1.16 1997/09/18 01:08:10 brianp Exp $ */
+/* $Id: tritemp.h,v 1.2.2.1 2000/06/18 22:48:20 brianp Exp $ */
 
 /*
  * Mesa 3-D graphics library
- * Version:  2.4
- * Copyright (C) 1995-1997  Brian Paul
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Library General Public
- * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Library General Public License for more details.
- *
- * You should have received a copy of the GNU Library General Public
- * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
- */
-
-
-/*
- * $Log: tritemp.h,v $
- * Revision 1.16  1997/09/18 01:08:10  brianp
- * fixed S_SCALE / T_SCALE mix-up
- *
- * Revision 1.15  1997/08/22 01:53:03  brianp
- * another attempt at fixing under/overflow errors
- *
- * Revision 1.14  1997/08/13 02:10:13  brianp
- * added code to prevent over/underflow (Guido Jansen, Magnus Lundin)
- *
- * Revision 1.13  1997/06/20 02:52:49  brianp
- * changed color components from GLfixed to GLubyte
- *
- * Revision 1.12  1997/03/14 00:25:02  brianp
- * fixed unitialized memory read, contributed by Tom Schmidt
- *
- * Revision 1.11  1997/02/09 18:51:10  brianp
- * fixed typo in texture R interpolation code
- *
- * Revision 1.10  1996/12/20 23:12:23  brianp
- * another attempt at preventing color interpolation over/underflow
- *
- * Revision 1.9  1996/12/18 20:38:25  brianp
- * commented out unused zp declaration
- *
- * Revision 1.8  1996/12/12 22:37:49  brianp
- * projective textures didn't work right
- *
- * Revision 1.7  1996/11/02 06:17:37  brianp
- * fixed some float/int roundoff and over/underflow errors (hopefully)
- *
- * Revision 1.6  1996/10/01 04:13:09  brianp
- * fixed Z interpolation for >16-bit depth buffer
- * added color underflow error check
- *
- * Revision 1.5  1996/09/27 01:32:59  brianp
- * removed unused variables
- *
- * Revision 1.4  1996/09/18 01:03:43  brianp
- * tightened threshold for culling by area
- *
- * Revision 1.3  1996/09/15 14:19:16  brianp
- * now use GLframebuffer and GLvisual
- *
- * Revision 1.2  1996/09/14 06:41:38  brianp
- * perspective correct texture code wasn't sub-pixel accurate (Doug Rabson)
- *
- * Revision 1.1  1996/09/13 01:38:16  brianp
- * Initial revision
- *
+ * Version:  3.1
+ * 
+ * Copyright (C) 1999  Brian Paul   All Rights Reserved.
+ * 
+ * Permission is hereby granted, free of charge, to any person obtaining a
+ * copy of this software and associated documentation files (the "Software"),
+ * to deal in the Software without restriction, including without limitation
+ * the rights to use, copy, modify, merge, publish, distribute, sublicense,
+ * and/or sell copies of the Software, and to permit persons to whom the
+ * Software is furnished to do so, subject to the following conditions:
+ * 
+ * The above copyright notice and this permission notice shall be included
+ * in all copies or substantial portions of the Software.
+ * 
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
+ * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
+ * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
+ * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
+ * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
  */
 
 
@@ -84,14 +34,14 @@
  * must be interplated across the triangle:
  *    INTERP_Z      - if defined, interpolate Z values
  *    INTERP_RGB    - if defined, interpolate RGB values
+ *    INTERP_SPEC   - if defined, interpolate specular RGB values
  *    INTERP_ALPHA  - if defined, interpolate Alpha values
  *    INTERP_INDEX  - if defined, interpolate color index values
- *    INTERP_ST     - if defined, interpolate integer ST texcoords
+ *    INTERP_INT_ST - if defined, interpolate integer ST texcoords
  *                         (fast, simple 2-D texture mapping)
- *    INTERP_STW    - if defined, interpolate float ST texcoords and W
- *                         (2-D texture maps with perspective correction)
- *    INTERP_UV     - if defined, interpolate float UV texcoords too
- *                         (for 3-D, 4-D? texture maps)
+ *    INTERP_STUV   - if defined, interpolate set 0 float STRQ texcoords
+ *                         NOTE:  OpenGL STRQ = Mesa STUV (R was taken for red)
+ *    INTERP_STUV1  - if defined, interpolate set 1 float STRQ texcoords
  *
  * When one can directly address pixels in the color buffer the following
  * macros can be defined and used to compute pixel addresses during
@@ -122,7 +72,7 @@
 /*void triangle( GLcontext *ctx, GLuint v0, GLuint v1, GLuint v2, GLuint pv )*/
 {
    typedef struct {
-        GLint v0, v1;     /* Y(v0) < Y(v1) */
+        GLint v0, v1;   /* Y(v0) < Y(v1) */
 	GLfloat dx;	/* X(v1) - X(v0) */
 	GLfloat dy;	/* Y(v1) - Y(v0) */
 	GLfixed fdxdy;	/* dx/dy in fixed-point */
@@ -137,12 +87,13 @@
    EdgeT eMaj, eTop, eBot;
    GLfloat oneOverArea;
    int vMin, vMid, vMax;       /* vertex indexes:  Y(vMin)<=Y(vMid)<=Y(vMax) */
+   float bf = ctx->backface_sign;
 
    /* find the order of the 3 vertices along the Y axis */
    {
-      GLfloat y0 = VB->Win[v0][1];
-      GLfloat y1 = VB->Win[v1][1];
-      GLfloat y2 = VB->Win[v2][1];
+      GLfloat y0 = VB->Win.data[v0][1];
+      GLfloat y1 = VB->Win.data[v1][1];
+      GLfloat y2 = VB->Win.data[v2][1];
 
       if (y0<=y1) {
 	 if (y1<=y2) {
@@ -152,15 +103,15 @@
 	    vMin = v2;   vMid = v0;   vMax = v1;   /* y2<=y0<=y1 */
 	 }
 	 else {
-	    vMin = v0;   vMid = v2;   vMax = v1;   /* y0<=y2<=y1 */
+	    vMin = v0;   vMid = v2;   vMax = v1;  bf = -bf; /* y0<=y2<=y1 */
 	 }
       }
       else {
 	 if (y0<=y2) {
-	    vMin = v1;   vMid = v0;   vMax = v2;   /* y1<=y0<=y2 */
+	    vMin = v1;   vMid = v0;   vMax = v2;  bf = -bf; /* y1<=y0<=y2 */
 	 }
 	 else if (y2<=y1) {
-	    vMin = v2;   vMid = v1;   vMax = v0;   /* y2<=y1<=y0 */
+	    vMin = v2;   vMid = v1;   vMax = v0;  bf = -bf; /* y2<=y1<=y0 */
 	 }
 	 else {
 	    vMin = v1;   vMid = v2;   vMax = v0;   /* y1<=y2<=y0 */
@@ -174,30 +125,39 @@
    eBot.v0 = vMin;   eBot.v1 = vMid;
 
    /* compute deltas for each edge:  vertex[v1] - vertex[v0] */
-   eMaj.dx = VB->Win[vMax][0] - VB->Win[vMin][0];
-   eMaj.dy = VB->Win[vMax][1] - VB->Win[vMin][1];
-   eTop.dx = VB->Win[vMax][0] - VB->Win[vMid][0];
-   eTop.dy = VB->Win[vMax][1] - VB->Win[vMid][1];
-   eBot.dx = VB->Win[vMid][0] - VB->Win[vMin][0];
-   eBot.dy = VB->Win[vMid][1] - VB->Win[vMin][1];
+   eMaj.dx = VB->Win.data[vMax][0] - VB->Win.data[vMin][0];
+   eMaj.dy = VB->Win.data[vMax][1] - VB->Win.data[vMin][1];
+   eTop.dx = VB->Win.data[vMax][0] - VB->Win.data[vMid][0];
+   eTop.dy = VB->Win.data[vMax][1] - VB->Win.data[vMid][1];
+   eBot.dx = VB->Win.data[vMid][0] - VB->Win.data[vMin][0];
+   eBot.dy = VB->Win.data[vMid][1] - VB->Win.data[vMin][1];
 
    /* compute oneOverArea */
    {
       GLfloat area = eMaj.dx * eBot.dy - eBot.dx * eMaj.dy;
-      if (area>-0.05f && area<0.05f) {
-         return;  /* very small; CULLED */
-      }
-      oneOverArea = 1.0F / area;
+
+      /* Do backface culling */
+      if (area * bf < 0.0)
+	 return;
+
+      if (area == 0.0F)
+         return;
+
+      /* check for very tiny triangle */
+      if (area * area < 0.0025F)  /* square it to ensure positive value */
+         oneOverArea = 1.0F / 0.0025F;  /* a close-enough value */
+      else
+         oneOverArea = 1.0F / area;
    }
 
    /* Edge setup.  For a triangle strip these could be reused... */
    {
       /* fixed point Y coordinates */
-      GLfixed vMin_fx = FloatToFixed(VB->Win[vMin][0] + 0.5F);
-      GLfixed vMin_fy = FloatToFixed(VB->Win[vMin][1] - 0.5F);
-      GLfixed vMid_fx = FloatToFixed(VB->Win[vMid][0] + 0.5F);
-      GLfixed vMid_fy = FloatToFixed(VB->Win[vMid][1] - 0.5F);
-      GLfixed vMax_fy = FloatToFixed(VB->Win[vMax][1] - 0.5F);
+      GLfixed vMin_fx = FloatToFixed(VB->Win.data[vMin][0] + 0.5F);
+      GLfixed vMin_fy = FloatToFixed(VB->Win.data[vMin][1] - 0.5F);
+      GLfixed vMid_fx = FloatToFixed(VB->Win.data[vMid][0] + 0.5F);
+      GLfixed vMid_fy = FloatToFixed(VB->Win.data[vMid][1] - 0.5F);
+      GLfixed vMax_fy = FloatToFixed(VB->Win.data[vMax][1] - 0.5F);
 
       eMaj.fsy = FixedCeil(vMin_fy);
       eMaj.lines = FixedToInt(vMax_fy + FIXED_ONE - FIXED_EPSILON - eMaj.fsy);
@@ -277,25 +237,33 @@
       GLfloat dgdx, dgdy;      GLfixed fdgdx;
       GLfloat dbdx, dbdy;      GLfixed fdbdx;
 #endif
+#if INTERP_SPEC
+      GLfloat dsrdx, dsrdy;    GLfixed fdsrdx;
+      GLfloat dsgdx, dsgdy;    GLfixed fdsgdx;
+      GLfloat dsbdx, dsbdy;    GLfixed fdsbdx;
+#endif
 #if INTERP_ALPHA
       GLfloat dadx, dady;      GLfixed fdadx;
 #endif
 #if INTERP_INDEX
       GLfloat didx, didy;      GLfixed fdidx;
 #endif
-#if INTERP_ST
+#if INTERP_INT_ST
       GLfloat dsdx, dsdy;      GLfixed fdsdx;
       GLfloat dtdx, dtdy;      GLfixed fdtdx;
 #endif
-#if INTERP_STW
+#if INTERP_STUV
       GLfloat dsdx, dsdy;
       GLfloat dtdx, dtdy;
-      GLfloat dwdx, dwdy;
-#endif
-#if INTERP_UV
       GLfloat dudx, dudy;
       GLfloat dvdx, dvdy;
 #endif
+#if INTERP_STUV1
+      GLfloat ds1dx, ds1dy;
+      GLfloat dt1dx, dt1dy;
+      GLfloat du1dx, du1dy;
+      GLfloat dv1dx, dv1dy;
+#endif
 
       /*
        * Execute user-supplied setup code
@@ -310,8 +278,8 @@
 #if INTERP_Z
       {
          GLfloat eMaj_dz, eBot_dz;
-         eMaj_dz = VB->Win[vMax][2] - VB->Win[vMin][2];
-         eBot_dz = VB->Win[vMid][2] - VB->Win[vMin][2];
+         eMaj_dz = VB->Win.data[vMax][2] - VB->Win.data[vMin][2];
+         eBot_dz = VB->Win.data[vMid][2] - VB->Win.data[vMin][2];
          dzdx = oneOverArea * (eMaj_dz * eBot.dy - eMaj.dy * eBot_dz);
          if (dzdx>DEPTH_SCALE || dzdx<-DEPTH_SCALE) {
             /* probably a sliver triangle */
@@ -331,34 +299,60 @@
 #if INTERP_RGB
       {
          GLfloat eMaj_dr, eBot_dr;
-         eMaj_dr = (GLint) VB->Color[vMax][0] - (GLint) VB->Color[vMin][0];
-         eBot_dr = (GLint) VB->Color[vMid][0] - (GLint) VB->Color[vMin][0];
+         eMaj_dr = (GLint) VB->ColorPtr->data[vMax][0] - (GLint) VB->ColorPtr->data[vMin][0];
+         eBot_dr = (GLint) VB->ColorPtr->data[vMid][0] - (GLint) VB->ColorPtr->data[vMin][0];
          drdx = oneOverArea * (eMaj_dr * eBot.dy - eMaj.dy * eBot_dr);
          fdrdx = SignedFloatToFixed(drdx);
          drdy = oneOverArea * (eMaj.dx * eBot_dr - eMaj_dr * eBot.dx);
       }
       {
          GLfloat eMaj_dg, eBot_dg;
-         eMaj_dg = (GLint) VB->Color[vMax][1] - (GLint) VB->Color[vMin][1];
-	 eBot_dg = (GLint) VB->Color[vMid][1] - (GLint) VB->Color[vMin][1];
+         eMaj_dg = (GLint) VB->ColorPtr->data[vMax][1] - (GLint) VB->ColorPtr->data[vMin][1];
+	 eBot_dg = (GLint) VB->ColorPtr->data[vMid][1] - (GLint) VB->ColorPtr->data[vMin][1];
          dgdx = oneOverArea * (eMaj_dg * eBot.dy - eMaj.dy * eBot_dg);
          fdgdx = SignedFloatToFixed(dgdx);
          dgdy = oneOverArea * (eMaj.dx * eBot_dg - eMaj_dg * eBot.dx);
       }
       {
          GLfloat eMaj_db, eBot_db;
-         eMaj_db = (GLint) VB->Color[vMax][2] - (GLint) VB->Color[vMin][2];
-         eBot_db = (GLint) VB->Color[vMid][2] - (GLint) VB->Color[vMin][2];
+         eMaj_db = (GLint) VB->ColorPtr->data[vMax][2] - (GLint) VB->ColorPtr->data[vMin][2];
+         eBot_db = (GLint) VB->ColorPtr->data[vMid][2] - (GLint) VB->ColorPtr->data[vMin][2];
          dbdx = oneOverArea * (eMaj_db * eBot.dy - eMaj.dy * eBot_db);
          fdbdx = SignedFloatToFixed(dbdx);
 	 dbdy = oneOverArea * (eMaj.dx * eBot_db - eMaj_db * eBot.dx);
       }
 #endif
+#if INTERP_SPEC
+      {
+         GLfloat eMaj_dsr, eBot_dsr;
+         eMaj_dsr = (GLint) VB->Specular[vMax][0] - (GLint) VB->Specular[vMin][0];
+         eBot_dsr = (GLint) VB->Specular[vMid][0] - (GLint) VB->Specular[vMin][0];
+         dsrdx = oneOverArea * (eMaj_dsr * eBot.dy - eMaj.dy * eBot_dsr);
+         fdsrdx = SignedFloatToFixed(dsrdx);
+         dsrdy = oneOverArea * (eMaj.dx * eBot_dsr - eMaj_dsr * eBot.dx);
+      }
+      {
+         GLfloat eMaj_dsg, eBot_dsg;
+         eMaj_dsg = (GLint) VB->Specular[vMax][1] - (GLint) VB->Specular[vMin][1];
+	 eBot_dsg = (GLint) VB->Specular[vMid][1] - (GLint) VB->Specular[vMin][1];
+         dsgdx = oneOverArea * (eMaj_dsg * eBot.dy - eMaj.dy * eBot_dsg);
+         fdsgdx = SignedFloatToFixed(dsgdx);
+         dsgdy = oneOverArea * (eMaj.dx * eBot_dsg - eMaj_dsg * eBot.dx);
+      }
+      {
+         GLfloat eMaj_dsb, eBot_dsb;
+         eMaj_dsb = (GLint) VB->Specular[vMax][2] - (GLint) VB->Specular[vMin][2];
+         eBot_dsb = (GLint) VB->Specular[vMid][2] - (GLint) VB->Specular[vMin][2];
+         dsbdx = oneOverArea * (eMaj_dsb * eBot.dy - eMaj.dy * eBot_dsb);
+         fdsbdx = SignedFloatToFixed(dsbdx);
+	 dsbdy = oneOverArea * (eMaj.dx * eBot_dsb - eMaj_dsb * eBot.dx);
+      }
+#endif
 #if INTERP_ALPHA
       {
          GLfloat eMaj_da, eBot_da;
-         eMaj_da = (GLint) VB->Color[vMax][3] - (GLint) VB->Color[vMin][3];
-         eBot_da = (GLint) VB->Color[vMid][3] - (GLint) VB->Color[vMin][3];
+         eMaj_da = (GLint) VB->ColorPtr->data[vMax][3] - (GLint) VB->ColorPtr->data[vMin][3];
+         eBot_da = (GLint) VB->ColorPtr->data[vMid][3] - (GLint) VB->ColorPtr->data[vMin][3];
          dadx = oneOverArea * (eMaj_da * eBot.dy - eMaj.dy * eBot_da);
          fdadx = SignedFloatToFixed(dadx);
          dady = oneOverArea * (eMaj.dx * eBot_da - eMaj_da * eBot.dx);
@@ -367,69 +361,143 @@
 #if INTERP_INDEX
       {
          GLfloat eMaj_di, eBot_di;
-         eMaj_di = (GLint) VB->Index[vMax] - (GLint) VB->Index[vMin];
-         eBot_di = (GLint) VB->Index[vMid] - (GLint) VB->Index[vMin];
+         eMaj_di = (GLint) VB->IndexPtr->data[vMax] - (GLint) VB->IndexPtr->data[vMin];
+         eBot_di = (GLint) VB->IndexPtr->data[vMid] - (GLint) VB->IndexPtr->data[vMin];
          didx = oneOverArea * (eMaj_di * eBot.dy - eMaj.dy * eBot_di);
          fdidx = SignedFloatToFixed(didx);
          didy = oneOverArea * (eMaj.dx * eBot_di - eMaj_di * eBot.dx);
       }
 #endif
-#if INTERP_ST
+#if INTERP_INT_ST
       {
          GLfloat eMaj_ds, eBot_ds;
-         eMaj_ds = (VB->TexCoord[vMax][0] - VB->TexCoord[vMin][0]) * S_SCALE;
-         eBot_ds = (VB->TexCoord[vMid][0] - VB->TexCoord[vMin][0]) * S_SCALE;
+         eMaj_ds = (VB->TexCoordPtr[0]->data[vMax][0] - VB->TexCoordPtr[0]->data[vMin][0]) * S_SCALE;
+         eBot_ds = (VB->TexCoordPtr[0]->data[vMid][0] - VB->TexCoordPtr[0]->data[vMin][0]) * S_SCALE;
          dsdx = oneOverArea * (eMaj_ds * eBot.dy - eMaj.dy * eBot_ds);
          fdsdx = SignedFloatToFixed(dsdx);
          dsdy = oneOverArea * (eMaj.dx * eBot_ds - eMaj_ds * eBot.dx);
       }
+      if (VB->TexCoordPtr[0]->size > 1)
       {
          GLfloat eMaj_dt, eBot_dt;
-         eMaj_dt = (VB->TexCoord[vMax][1] - VB->TexCoord[vMin][1]) * T_SCALE;
-         eBot_dt = (VB->TexCoord[vMid][1] - VB->TexCoord[vMin][1]) * T_SCALE;
+         eMaj_dt = (VB->TexCoordPtr[0]->data[vMax][1] - VB->TexCoordPtr[0]->data[vMin][1]) * T_SCALE;
+         eBot_dt = (VB->TexCoordPtr[0]->data[vMid][1] - VB->TexCoordPtr[0]->data[vMin][1]) * T_SCALE;
          dtdx = oneOverArea * (eMaj_dt * eBot.dy - eMaj.dy * eBot_dt);
          fdtdx = SignedFloatToFixed(dtdx);
          dtdy = oneOverArea * (eMaj.dx * eBot_dt - eMaj_dt * eBot.dx);
+      } else {
+         dtdx = 0;
+         fdtdx = SignedFloatToFixed(dtdx);
+         dtdy = 0;
       }
+
 #endif
-#if INTERP_STW
+#if INTERP_STUV
       {
-         GLfloat wMax = 1.0F / VB->Clip[vMax][3];
-         GLfloat wMin = 1.0F / VB->Clip[vMin][3];
-         GLfloat wMid = 1.0F / VB->Clip[vMid][3];
-         GLfloat eMaj_dw, eBot_dw;
+         GLfloat wMax = VB->Win.data[vMax][3];
+         GLfloat wMin = VB->Win.data[vMin][3];
+         GLfloat wMid = VB->Win.data[vMid][3];
          GLfloat eMaj_ds, eBot_ds;
          GLfloat eMaj_dt, eBot_dt;
-#if INTERP_UV
          GLfloat eMaj_du, eBot_du;
          GLfloat eMaj_dv, eBot_dv;
-#endif
-         eMaj_dw = wMax - wMin;
-         eBot_dw = wMid - wMin;
-         dwdx = oneOverArea * (eMaj_dw * eBot.dy - eMaj.dy * eBot_dw);
-         dwdy = oneOverArea * (eMaj.dx * eBot_dw - eMaj_dw * eBot.dx);
 
-         eMaj_ds = VB->TexCoord[vMax][0]*wMax - VB->TexCoord[vMin][0]*wMin;
-         eBot_ds = VB->TexCoord[vMid][0]*wMid - VB->TexCoord[vMin][0]*wMin;
+         eMaj_ds = VB->TexCoordPtr[0]->data[vMax][0]*wMax - VB->TexCoordPtr[0]->data[vMin][0]*wMin;
+         eBot_ds = VB->TexCoordPtr[0]->data[vMid][0]*wMid - VB->TexCoordPtr[0]->data[vMin][0]*wMin;
          dsdx = oneOverArea * (eMaj_ds * eBot.dy - eMaj.dy * eBot_ds);
          dsdy = oneOverArea * (eMaj.dx * eBot_ds - eMaj_ds * eBot.dx);
 
-         eMaj_dt = VB->TexCoord[vMax][1]*wMax - VB->TexCoord[vMin][1]*wMin;
-         eBot_dt = VB->TexCoord[vMid][1]*wMid - VB->TexCoord[vMin][1]*wMin;
-         dtdx = oneOverArea * (eMaj_dt * eBot.dy - eMaj.dy * eBot_dt);
-         dtdy = oneOverArea * (eMaj.dx * eBot_dt - eMaj_dt * eBot.dx);
-#if INTERP_UV
-         eMaj_du = VB->TexCoord[vMax][2]*wMax - VB->TexCoord[vMin][2]*wMin;
-         eBot_du = VB->TexCoord[vMid][2]*wMid - VB->TexCoord[vMin][2]*wMin;
-         dudx = oneOverArea * (eMaj_du * eBot.dy - eMaj.dy * eBot_du);
-         dudy = oneOverArea * (eMaj.dx * eBot_du - eMaj_du * eBot.dx);
 
-         /* Note: don't divide V component by W */
-         eMaj_dv = VB->TexCoord[vMax][3] - VB->TexCoord[vMin][3];
-         eBot_dv = VB->TexCoord[vMid][3] - VB->TexCoord[vMin][3];
-         dvdx = oneOverArea * (eMaj_dv * eBot.dy - eMaj.dy * eBot_dv);
-         dvdy = oneOverArea * (eMaj.dx * eBot_dv - eMaj_dv * eBot.dx);
+	 if (VB->TexCoordPtr[0]->size > 1)
+	 {
+	    eMaj_dt = VB->TexCoordPtr[0]->data[vMax][1]*wMax - VB->TexCoordPtr[0]->data[vMin][1]*wMin;
+	    eBot_dt = VB->TexCoordPtr[0]->data[vMid][1]*wMid - VB->TexCoordPtr[0]->data[vMin][1]*wMin;
+	    dtdx = oneOverArea * (eMaj_dt * eBot.dy - eMaj.dy * eBot_dt);
+	    dtdy = oneOverArea * (eMaj.dx * eBot_dt - eMaj_dt * eBot.dx);
+	 } else {
+	    dtdx = 0;
+	    dtdy = 0; 
+	 }
+
+	 if (VB->TexCoordPtr[0]->size > 2)
+	 {
+	    eMaj_du = VB->TexCoordPtr[0]->data[vMax][2]*wMax - VB->TexCoordPtr[0]->data[vMin][2]*wMin;
+	    eBot_du = VB->TexCoordPtr[0]->data[vMid][2]*wMid - VB->TexCoordPtr[0]->data[vMin][2]*wMin;
+	    dudx = oneOverArea * (eMaj_du * eBot.dy - eMaj.dy * eBot_du);
+	    dudy = oneOverArea * (eMaj.dx * eBot_du - eMaj_du * eBot.dx);
+	 } else {
+	    dudx = 0;
+	    dudy = 0;
+	 }
+
+	 if (VB->TexCoordPtr[0]->size > 3)
+	 {
+	    eMaj_dv = VB->TexCoordPtr[0]->data[vMax][3]*wMax - VB->TexCoordPtr[0]->data[vMin][3]*wMin;
+	    eBot_dv = VB->TexCoordPtr[0]->data[vMid][3]*wMid - VB->TexCoordPtr[0]->data[vMin][3]*wMin;
+	    dvdx = oneOverArea * (eMaj_dv * eBot.dy - eMaj.dy * eBot_dv);
+	    dvdy = oneOverArea * (eMaj.dx * eBot_dv - eMaj_dv * eBot.dx);
+	 } else {
+	    eMaj_dv = wMax - wMin;
+	    eBot_dv = wMid - wMin;
+	    dvdx = oneOverArea * (eMaj_dv * eBot.dy - eMaj.dy * eBot_dv);
+	    dvdy = oneOverArea * (eMaj.dx * eBot_dv - eMaj_dv * eBot.dx);
+	 }
+      }
 #endif
+#if INTERP_STUV1
+      {
+         GLfloat wMax = VB->Win.data[vMax][3];
+         GLfloat wMin = VB->Win.data[vMin][3];
+         GLfloat wMid = VB->Win.data[vMid][3];
+         GLfloat eMaj_ds, eBot_ds;
+         GLfloat eMaj_dt, eBot_dt;
+         GLfloat eMaj_du, eBot_du;
+         GLfloat eMaj_dv, eBot_dv;
+         eMaj_ds = VB->TexCoordPtr[1]->data[vMax][0]*wMax - VB->TexCoordPtr[1]->data[vMin][0]*wMin;
+         eBot_ds = VB->TexCoordPtr[1]->data[vMid][0]*wMid - VB->TexCoordPtr[1]->data[vMin][0]*wMin;
+         ds1dx = oneOverArea * (eMaj_ds * eBot.dy - eMaj.dy * eBot_ds);
+         ds1dy = oneOverArea * (eMaj.dx * eBot_ds - eMaj_ds * eBot.dx);
+
+	 if (VB->TexCoordPtr[1]->size > 1)
+	 {
+	    eMaj_dt = VB->TexCoordPtr[1]->data[vMax][1]*wMax - VB->TexCoordPtr[1]->data[vMin][1]*wMin;
+	    eBot_dt = VB->TexCoordPtr[1]->data[vMid][1]*wMid - VB->TexCoordPtr[1]->data[vMin][1]*wMin;
+	    dt1dx = oneOverArea * (eMaj_dt * eBot.dy - eMaj.dy * eBot_dt);
+	    dt1dy = oneOverArea * (eMaj.dx * eBot_dt - eMaj_dt * eBot.dx);
+	 }
+	 else
+	 {
+	    dt1dx = 0;
+	    dt1dy = 0;
+	 }
+
+	 if (VB->TexCoordPtr[1]->size > 2)
+	 {
+	    eMaj_du = VB->TexCoordPtr[1]->data[vMax][2]*wMax - VB->TexCoordPtr[1]->data[vMin][2]*wMin;
+	    eBot_du = VB->TexCoordPtr[1]->data[vMid][2]*wMid - VB->TexCoordPtr[1]->data[vMin][2]*wMin;
+	    du1dx = oneOverArea * (eMaj_du * eBot.dy - eMaj.dy * eBot_du);
+	    du1dy = oneOverArea * (eMaj.dx * eBot_du - eMaj_du * eBot.dx);
+	 }
+	 else
+	 {
+	    du1dx = 0;
+	    du1dy = 0;
+	 }
+
+	 if (VB->TexCoordPtr[1]->size > 3)
+	 {
+	    eMaj_dv = VB->TexCoordPtr[1]->data[vMax][3]*wMax - VB->TexCoordPtr[1]->data[vMin][3]*wMin;
+	    eBot_dv = VB->TexCoordPtr[1]->data[vMid][3]*wMid - VB->TexCoordPtr[1]->data[vMin][3]*wMin;
+	    dv1dx = oneOverArea * (eMaj_dv * eBot.dy - eMaj.dy * eBot_dv);
+	    dv1dy = oneOverArea * (eMaj.dx * eBot_dv - eMaj_dv * eBot.dx);
+	 }
+	 else
+	 {
+	    eMaj_dv = wMax - wMin;
+	    eBot_dv = wMid - wMin;
+	    dv1dx = oneOverArea * (eMaj_dv * eBot.dy - eMaj.dy * eBot_dv);
+	    dv1dy = oneOverArea * (eMaj.dx * eBot_dv - eMaj_dv * eBot.dx);
+	 }
       }
 #endif
 
@@ -504,25 +572,33 @@
          GLfixed fg, fdgOuter, fdgInner;
          GLfixed fb, fdbOuter, fdbInner;
 #endif
+#if INTERP_SPEC
+         GLfixed fsr, fdsrOuter, fdsrInner;
+         GLfixed fsg, fdsgOuter, fdsgInner;
+         GLfixed fsb, fdsbOuter, fdsbInner;
+#endif
 #if INTERP_ALPHA
          GLfixed fa, fdaOuter, fdaInner;
 #endif
 #if INTERP_INDEX
          GLfixed fi, fdiOuter, fdiInner;
 #endif
-#if INTERP_ST
+#if INTERP_INT_ST
          GLfixed fs, fdsOuter, fdsInner;
          GLfixed ft, fdtOuter, fdtInner;
 #endif
-#if INTERP_STW
+#if INTERP_STUV
          GLfloat sLeft, dsOuter, dsInner;
          GLfloat tLeft, dtOuter, dtInner;
-         GLfloat wLeft, dwOuter, dwInner;
-#endif
-#if INTERP_UV
          GLfloat uLeft, duOuter, duInner;
          GLfloat vLeft, dvOuter, dvInner;
 #endif
+#if INTERP_STUV1
+         GLfloat s1Left, ds1Outer, ds1Inner;
+         GLfloat t1Left, dt1Outer, dt1Inner;
+         GLfloat u1Left, du1Outer, du1Inner;
+         GLfloat v1Left, dv1Outer, dv1Inner;
+#endif
 
          for (subTriangle=0; subTriangle<=1; subTriangle++) {
             EdgeT *eLeft, *eRight;
@@ -576,19 +652,23 @@
                fdError = fdxOuter - fdxLeftEdge + FIXED_ONE;
                idxOuter = FixedToInt(fdxOuter);
                dxOuter = (float) idxOuter;
+               (void) dxOuter;
 
                fy = eLeft->fsy;
                iy = FixedToInt(fy);
 
                adjx = (float)(fx - eLeft->fx0);  /* SCALED! */
                adjy = eLeft->adjy;		 /* SCALED! */
+               (void) adjx;  /* silence compiler warnings */
+               (void) adjy;  /* silence compiler warnings */
 
                vLower = eLeft->v0;
+               (void) vLower;  /* silence compiler warnings */
 
 #ifdef PIXEL_ADDRESS
                {
                   pRow = PIXEL_ADDRESS( FixedToInt(fxLeftEdge), iy );
-                  dPRowOuter = -BYTES_PER_ROW + idxOuter * sizeof(PIXEL_TYPE);
+                  dPRowOuter = -((int)BYTES_PER_ROW) + idxOuter * sizeof(PIXEL_TYPE);
                   /* negative because Y=0 at bottom and increases upward */
                }
 #endif
@@ -605,7 +685,7 @@
 #if INTERP_Z
                {
                   GLfloat z0, tmp;
-                  z0 = VB->Win[vLower][2] + ctx->PolygonZoffset;
+                  z0 = VB->Win.data[vLower][2] + ctx->PolygonZoffset;
 #if DEPTH_BITS==16
                   /* interpolate fixed-pt values */
                   tmp = (z0 * FIXED_SCALE + dzdx * adjx + dzdy * adjy) + FIXED_HALF;
@@ -615,6 +695,7 @@
                      fz = MAX_GLUINT/2;
                   fdzOuter = SignedFloatToFixed(dzdy + dxOuter * dzdx);
 #else
+                  (void) tmp;
                   /* interpolate depth values exactly */
                   fz = (GLint) (z0 + dzdx*FixedToFloat(adjx) + dzdy*FixedToFloat(adjy));
                   fdzOuter = (GLint) (dzdy + dxOuter * dzdx);
@@ -624,60 +705,126 @@
                }
 #endif
 #if INTERP_RGB
-               fr = (GLfixed)(IntToFixed(VB->Color[vLower][0]) + drdx * adjx + drdy * adjy)
+               fr = (GLfixed)(IntToFixed(VB->ColorPtr->data[vLower][0]) + drdx * adjx + drdy * adjy)
                     + FIXED_HALF;
                fdrOuter = SignedFloatToFixed(drdy + dxOuter * drdx);
 
-               fg = (GLfixed)(IntToFixed(VB->Color[vLower][1]) + dgdx * adjx + dgdy * adjy)
+               fg = (GLfixed)(IntToFixed(VB->ColorPtr->data[vLower][1]) + dgdx * adjx + dgdy * adjy)
                     + FIXED_HALF;
                fdgOuter = SignedFloatToFixed(dgdy + dxOuter * dgdx);
 
-               fb = (GLfixed)(IntToFixed(VB->Color[vLower][2]) + dbdx * adjx + dbdy * adjy)
+               fb = (GLfixed)(IntToFixed(VB->ColorPtr->data[vLower][2]) + dbdx * adjx + dbdy * adjy)
                     + FIXED_HALF;
                fdbOuter = SignedFloatToFixed(dbdy + dxOuter * dbdx);
 #endif
+#if INTERP_SPEC
+               fsr = (GLfixed)(IntToFixed(VB->Specular[vLower][0]) + dsrdx * adjx + dsrdy * adjy)
+                    + FIXED_HALF;
+               fdsrOuter = SignedFloatToFixed(dsrdy + dxOuter * dsrdx);
+
+               fsg = (GLfixed)(IntToFixed(VB->Specular[vLower][1]) + dsgdx * adjx + dsgdy * adjy)
+                    + FIXED_HALF;
+               fdsgOuter = SignedFloatToFixed(dsgdy + dxOuter * dsgdx);
+
+               fsb = (GLfixed)(IntToFixed(VB->Specular[vLower][2]) + dsbdx * adjx + dsbdy * adjy)
+                    + FIXED_HALF;
+               fdsbOuter = SignedFloatToFixed(dsbdy + dxOuter * dsbdx);
+#endif
 #if INTERP_ALPHA
-               fa = (GLfixed)(IntToFixed(VB->Color[vLower][3]) + dadx * adjx + dady * adjy)
+               fa = (GLfixed)(IntToFixed(VB->ColorPtr->data[vLower][3]) + dadx * adjx + dady * adjy)
                     + FIXED_HALF;
                fdaOuter = SignedFloatToFixed(dady + dxOuter * dadx);
 #endif
 #if INTERP_INDEX
-               fi = (GLfixed)(VB->Index[vLower] * FIXED_SCALE + didx * adjx
+               fi = (GLfixed)(VB->IndexPtr->data[vLower] * FIXED_SCALE + didx * adjx
                               + didy * adjy) + FIXED_HALF;
                fdiOuter = SignedFloatToFixed(didy + dxOuter * didx);
 #endif
-#if INTERP_ST
+#if INTERP_INT_ST
                {
                   GLfloat s0, t0;
-                  s0 = VB->TexCoord[vLower][0] * S_SCALE;
+                  s0 = VB->TexCoordPtr[0]->data[vLower][0] * S_SCALE;
                   fs = (GLfixed)(s0 * FIXED_SCALE + dsdx * adjx + dsdy * adjy) + FIXED_HALF;
                   fdsOuter = SignedFloatToFixed(dsdy + dxOuter * dsdx);
-                  t0 = VB->TexCoord[vLower][1] * T_SCALE;
-                  ft = (GLfixed)(t0 * FIXED_SCALE + dtdx * adjx + dtdy * adjy) + FIXED_HALF;
-                  fdtOuter = SignedFloatToFixed(dtdy + dxOuter * dtdx);
-               }
+
+		  if (VB->TexCoordPtr[0]->size > 1)
+		  {
+		     t0 = VB->TexCoordPtr[0]->data[vLower][1] * T_SCALE;
+		     ft = (GLfixed)(t0 * FIXED_SCALE + dtdx * adjx + dtdy * adjy) + FIXED_HALF;
+		     fdtOuter = SignedFloatToFixed(dtdy + dxOuter * dtdx);
+		  } 
+		  else
+		  {
+		     t0 = 0;
+		     ft = (GLfixed) FIXED_HALF;
+		     fdtOuter = SignedFloatToFixed(0);
+		  }
+	       }
 #endif
-#if INTERP_STW
+#if INTERP_STUV
                {
-                  GLfloat w0 = 1.0F / VB->Clip[vLower][3];
+                  GLfloat invW = VB->Win.data[vLower][3];
                   GLfloat s0, t0, u0, v0;
-                  wLeft = w0 + (dwdx * adjx + dwdy * adjy) * (1.0F/FIXED_SCALE);
-		  dwOuter = dwdy + dxOuter * dwdx;
-                  s0 = VB->TexCoord[vLower][0] * w0;
+                  s0 = VB->TexCoordPtr[0]->data[vLower][0] * invW;
                   sLeft = s0 + (dsdx * adjx + dsdy * adjy) * (1.0F/FIXED_SCALE);
                   dsOuter = dsdy + dxOuter * dsdx;
-                  t0 = VB->TexCoord[vLower][1] * w0;
-                  tLeft = t0 + (dtdx * adjx + dtdy * adjy) * (1.0F/FIXED_SCALE);
-                  dtOuter = dtdy + dxOuter * dtdx;
-#if INTERP_UV
-                  u0 = VB->TexCoord[vLower][2] * w0;
-                  uLeft = u0 + (dudx * adjx + dudy * adjy) * (1.0F/FIXED_SCALE);
-                  duOuter = dudy + dxOuter * dudx;
-                  /* Note: don't divide V component by W */
-                  v0 = VB->TexCoord[vLower][3];
-                  vLeft = v0 + (dvdx * adjx + dvdy * adjy) * (1.0F/FIXED_SCALE);
-                  dvOuter = dvdy + dxOuter * dvdx;
+		  if (VB->TexCoordPtr[0]->size > 1)
+		  {		  
+		     t0 = VB->TexCoordPtr[0]->data[vLower][1] * invW;
+		     tLeft = t0 + (dtdx * adjx + dtdy * adjy) * (1.0F/FIXED_SCALE);
+		     dtOuter = dtdy + dxOuter * dtdx;
+		  } else {
+		     tLeft = dtOuter = 0;
+		  }
+		  if (VB->TexCoordPtr[0]->size > 2)
+		  {		  
+		     u0 = VB->TexCoordPtr[0]->data[vLower][2] * invW;
+		     uLeft = u0 + (dudx * adjx + dudy * adjy) * (1.0F/FIXED_SCALE);
+		     duOuter = dudy + dxOuter * dudx;
+		  } else {
+		     uLeft = duOuter = 0;
+		  }
+		  if (VB->TexCoordPtr[0]->size > 3)
+		  {		  
+		     v0 = VB->TexCoordPtr[0]->data[vLower][3] * invW;
+		  } else {
+		     v0 = invW;
+		  }
+		  vLeft = v0 + (dvdx * adjx + dvdy * adjy) * (1.0F/FIXED_SCALE);
+		  dvOuter = dvdy + dxOuter * dvdx;
+               }
 #endif
+#if INTERP_STUV1
+               {
+                  GLfloat invW = VB->Win.data[vLower][3];
+                  GLfloat s0, t0, u0, v0;
+                  s0 = VB->TexCoordPtr[1]->data[vLower][0] * invW;
+                  s1Left = s0 + (ds1dx * adjx + ds1dy * adjy) * (1.0F/FIXED_SCALE);
+                  ds1Outer = ds1dy + dxOuter * ds1dx;
+		  if (VB->TexCoordPtr[0]->size > 1)
+		  {		  
+		     t0 = VB->TexCoordPtr[1]->data[vLower][1] * invW;
+		     t1Left = t0 + (dt1dx * adjx + dt1dy * adjy) * (1.0F/FIXED_SCALE);
+		     dt1Outer = dt1dy + dxOuter * dt1dx;
+		  } else {
+		     t1Left = dt1Outer = 0;
+		  }
+		  if (VB->TexCoordPtr[0]->size > 2)
+		  {		  
+		     u0 = VB->TexCoordPtr[1]->data[vLower][2] * invW;
+		     u1Left = u0 + (du1dx * adjx + du1dy * adjy) * (1.0F/FIXED_SCALE);
+		     du1Outer = du1dy + dxOuter * du1dx;
+		  } else {
+		     u1Left = du1Outer = 0;
+		  }
+		  if (VB->TexCoordPtr[0]->size > 3)
+		  {		  
+		     v0 = VB->TexCoordPtr[1]->data[vLower][3] * invW;
+		  } else {
+		     v0 =  invW;
+		  }
+		  v1Left = v0 + (dv1dx * adjx + dv1dy * adjy) * (1.0F/FIXED_SCALE);
+		  dv1Outer = dv1dy + dxOuter * dv1dx;
                }
 #endif
 
@@ -707,24 +854,32 @@
             fdgInner = fdgOuter + fdgdx;
             fdbInner = fdbOuter + fdbdx;
 #endif
+#if INTERP_SPEC
+            fdsrInner = fdsrOuter + fdsrdx;
+            fdsgInner = fdsgOuter + fdsgdx;
+            fdsbInner = fdsbOuter + fdsbdx;
+#endif
 #if INTERP_ALPHA
             fdaInner = fdaOuter + fdadx;
 #endif
 #if INTERP_INDEX
             fdiInner = fdiOuter + fdidx;
 #endif
-#if INTERP_ST
+#if INTERP_INT_ST
             fdsInner = fdsOuter + fdsdx;
             fdtInner = fdtOuter + fdtdx;
 #endif
-#if INTERP_STW
-	    dwInner = dwOuter + dwdx;
+#if INTERP_STUV
 	    dsInner = dsOuter + dsdx;
 	    dtInner = dtOuter + dtdx;
-#if INTERP_UV
 	    duInner = duOuter + dudx;
 	    dvInner = dvOuter + dvdx;
 #endif
+#if INTERP_STUV1
+	    ds1Inner = ds1Outer + ds1dx;
+	    dt1Inner = dt1Outer + dt1dx;
+	    du1Inner = du1Outer + du1dx;
+	    dv1Inner = dv1Outer + dv1dx;
 #endif
 
             while (lines>0) {
@@ -737,20 +892,23 @@
 #if INTERP_RGB
                GLfixed ffr = fr,  ffg = fg,  ffb = fb;
 #endif
+#if INTERP_SPEC
+               GLfixed ffsr = fsr,  ffsg = fsg,  ffsb = fsb;
+#endif
 #if INTERP_ALPHA
                GLfixed ffa = fa;
 #endif
 #if INTERP_INDEX
                GLfixed ffi = fi;
 #endif
-#if INTERP_ST
+#if INTERP_INT_ST
                GLfixed ffs = fs,  fft = ft;
 #endif
-#if INTERP_STW
-               GLfloat ss = sLeft,  tt = tLeft,  ww = wLeft;
+#if INTERP_STUV
+               GLfloat ss = sLeft, tt = tLeft, uu = uLeft, vv = vLeft;
 #endif
-#if INTERP_UV
-               GLfloat uu = uLeft,  vv = vLeft;
+#if INTERP_STUV1
+               GLfloat ss1 = s1Left, tt1 = t1Left, uu1 = u1Left, vv1 = v1Left;
 #endif
                GLint left = FixedToInt(fxLeftEdge);
                GLint right = FixedToInt(fxRightEdge);
@@ -769,6 +927,20 @@
                   if (ffb<0) ffb = 0;
                }
 #endif
+#if INTERP_SPEC
+               {
+                  /* need this to accomodate round-off errors */
+                  GLfixed ffsrend = ffsr+(right-left-1)*fdsrdx;
+                  GLfixed ffsgend = ffsg+(right-left-1)*fdsgdx;
+                  GLfixed ffsbend = ffsb+(right-left-1)*fdsbdx;
+                  if (ffsrend<0) ffsr -= ffsrend;
+                  if (ffsgend<0) ffsg -= ffsgend;
+                  if (ffsbend<0) ffsb -= ffsbend;
+                  if (ffsr<0) ffsr = 0;
+                  if (ffsg<0) ffsg = 0;
+                  if (ffsb<0) ffsb = 0;
+               }
+#endif
 #if INTERP_ALPHA
                {
                   GLfixed ffaend = ffa+(right-left-1)*fdadx;
@@ -808,23 +980,29 @@
 #if INTERP_RGB
                   fr += fdrOuter;   fg += fdgOuter;   fb += fdbOuter;
 #endif
+#if INTERP_SPEC
+                  fsr += fdsrOuter;   fsg += fdsgOuter;   fsb += fdsbOuter;
+#endif
 #if INTERP_ALPHA
                   fa += fdaOuter;
 #endif
 #if INTERP_INDEX
                   fi += fdiOuter;
 #endif
-#if INTERP_ST
+#if INTERP_INT_ST
                   fs += fdsOuter;   ft += fdtOuter;
 #endif
-#if INTERP_STW
+#if INTERP_STUV
 		  sLeft += dsOuter;
 		  tLeft += dtOuter;
-		  wLeft += dwOuter;
-#endif
-#if INTERP_UV
 		  uLeft += duOuter;
 		  vLeft += dvOuter;
+#endif
+#if INTERP_STUV1
+		  s1Left += ds1Outer;
+		  t1Left += dt1Outer;
+		  u1Left += du1Outer;
+		  v1Left += dv1Outer;
 #endif
                }
                else {
@@ -838,23 +1016,29 @@
 #if INTERP_RGB
                   fr += fdrInner;   fg += fdgInner;   fb += fdbInner;
 #endif
+#if INTERP_SPEC
+                  fsr += fdsrInner;   fsg += fdsgInner;   fsb += fdsbInner;
+#endif
 #if INTERP_ALPHA
                   fa += fdaInner;
 #endif
 #if INTERP_INDEX
                   fi += fdiInner;
 #endif
-#if INTERP_ST
+#if INTERP_INT_ST
                   fs += fdsInner;   ft += fdtInner;
 #endif
-#if INTERP_STW
+#if INTERP_STUV
 		  sLeft += dsInner;
 		  tLeft += dtInner;
-		  wLeft += dwInner;
-#endif
-#if INTERP_UV
 		  uLeft += duInner;
 		  vLeft += dvInner;
+#endif
+#if INTERP_STUV1
+		  s1Left += ds1Inner;
+		  t1Left += dt1Inner;
+		  u1Left += du1Inner;
+		  v1Left += dv1Inner;
 #endif
                }
             } /*while lines>0*/
@@ -874,11 +1058,12 @@
 
 #undef INTERP_Z
 #undef INTERP_RGB
+#undef INTERP_SPEC
 #undef INTERP_ALPHA
 #undef INTERP_INDEX
-#undef INTERP_ST
-#undef INTERP_STW
-#undef INTERP_UV
+#undef INTERP_INT_ST
+#undef INTERP_STUV
+#undef INTERP_STUV1
 
 #undef S_SCALE
 #undef T_SCALE
