@@ -1,5 +1,5 @@
 /* -*- mode: C; tab-width:8; c-basic-offset:8 -*- */
-/* $Id: types.h,v 1.18.2.9 2000/02/21 22:49:24 brianp Exp $ */
+/* $Id: types.h,v 1.18.2.10 2000/06/27 23:39:49 brianp Exp $ */
 
 /*
  * Mesa 3-D graphics library
@@ -1892,7 +1892,7 @@ struct gl_context {
 	GLmatrix ProjectionMatrix;
 	GLuint ProjectionStackDepth;
 	GLmatrix ProjectionStack[MAX_PROJECTION_STACK_DEPTH - 1];
-	GLfloat NearFarStack[MAX_PROJECTION_STACK_DEPTH - 1][2];
+	GLfloat NearFarStack[MAX_PROJECTION_STACK_DEPTH][2];
 
         /* Combined modelview and projection matrix */
         GLmatrix ModelProjectMatrix;
