@@ -1,4 +1,4 @@
-/* $Id: vbcull.c,v 1.7 1999/11/08 15:28:58 brianp Exp $ */
+/* $Id: vbcull.c,v 1.7.2.1 2000/06/29 14:14:14 brianp Exp $ */
 
 /*
  * Mesa 3-D graphics library
@@ -96,7 +96,7 @@ static GLuint gl_cull_points( struct vertex_buffer *VB,
 
    /* This is pretty pointless.  (arf arf)
     */
-   for (i=start+1;i<count;i++) {
+   for (i = start; i < count; i++) {
       if (clipmask[i] == 0) 
 	 cullmask[i] = VERT_FACE_FRONT | PRIM_FACE_FRONT;
       else {
